#include "init_square.h"

// << move DOWN on visual board, >> move UP on visual board.


/********************************************************************************
 * functions to ease the process of finding the initial square of a moving piece :
********************************************************************************/

//from wikipedia http://en.wikipedia.org/wiki/Binary_logarithm
//to find the last bit set
int flsll(Bitboard n) {
    if (n == 0) return 0;
    int pos = 1;
    if (n >= (Bitboard)1<<32) {
        n >>= 32; pos += 32; }
        if (n >= 1<<16) { n >>= 16; pos += 16; }
        if (n >= 1<< 8) { n >>=  8; pos +=  8; }
        if (n >= 1<< 4) { n >>=  4; pos +=  4; }
        if (n >= 1<< 2) { n >>=  2; pos +=  2; }
        if (n >= 1<< 1) {           pos +=  1; }
        return pos;
}

// when sliders move, we have to check when it hits chessboard's boundaries
// this function return the right bitboard for a given initial square and direction
Bitboard getBoundaryBit(Bitboard square, int leftShift, int sizeShift) {

    if (leftShift) {
    	// left shift : moving down
        if (sizeShift == 8) return ranks[0];
        if (sizeShift == 1) return files[7];

        int firstBit_sq = ffsll(square) - 1;
        int file_sq = firstBit_sq % 8;
        int rank_sq = 7 - firstBit_sq / 8;
        if (sizeShift == 7) {
            // we are moving down left (see above)
            if (file_sq <= rank_sq) return files[0];
            return ranks[0];
        }
        else {
            // ..moving down right
            int mRank_sq = 7 - rank_sq;
            if (file_sq >= mRank_sq) return files[7];
            return ranks[0];
        }
    }

    else {
        // shift right : moving up
        if (sizeShift == 8) return ranks[7];
        if (sizeShift == 1) return files[0];

        Bitboard firstBit_sq = ffsll(square) - 1;
        int file_sq = firstBit_sq % 8;
        int rank_sq = 7 - firstBit_sq / 8;
        if (sizeShift == 7) {
            // we are moving up right
            if (file_sq <= rank_sq) return ranks[7];
            return files[7];
        }
        else {
            // ..and moving up left
            int mRank_sq = 7 - rank_sq;
            if (file_sq >= mRank_sq) return ranks[7];
            return files[0];
        }
    }

    return 0; // never reached
}

// this function will take a destination square in parameter,
// and slide toward the indicated direction until it meets some corresponding slider
// (rook, bishop, queen) ; it can fail, and return #1111... (all 1's)
Bitboard tryOneWay(BoardState* bs, Bitboard destSquare, Bitboard piecesPos, int leftShift, int sizeShift) {

    // first compute boundary bit for this way
    Bitboard bb = getBoundaryBit(destSquare,leftShift,sizeShift);
    if (destSquare & bb) return 0xFFFFFFFFFFFFFFFF; // means this way doesn't work

    Bitboard moving_ds = 0;
    if (leftShift) {
        moving_ds = destSquare << sizeShift;
        while ( moving_ds && ! (moving_ds & bb) && ! (moving_ds & bs->fullBoard) )
            moving_ds <<= sizeShift;
    }

    else {
        moving_ds = destSquare >> sizeShift;
        while ( moving_ds && ! (moving_ds & bb) && ! (moving_ds & bs->fullBoard) )
            moving_ds >>= sizeShift;
    }

	// finally do a quick "king in check" test to validate the move :
    if ( (moving_ds & piecesPos) && !checkCheck(bs, moving_ds, destSquare) ) return moving_ds;
    return 0xFFFFFFFFFFFFFFFF; // means this way doesn't work
}

// this function simply calls the one above, for two directions :
Bitboard tryTwoWays(BoardState* bs, Bitboard destSquare, Bitboard piecesPos,
    int leftFirstShift, int sizeFirstShift, int leftNextShift, int sizeNextShift) {

    // try first way :
    Bitboard moving_ds = tryOneWay(bs, destSquare, piecesPos, leftFirstShift, sizeFirstShift);
    if ( ~moving_ds ) return moving_ds;

    // bad luck, try other way :
    return (tryOneWay(bs, destSquare, piecesPos, leftNextShift, sizeNextShift));
}


/*******************************************************************************
 * Main functions for finding initial square of some piece, from pawn to queen :
*******************************************************************************/

// pawns == easy. no promotions because already done in the main parser.
Bitboard getInitSquare_pawns(BoardState* bs, Bitboard destSquare, int take, int file_is) {
    Bitboard initSquare=0;

    if (take) {
        // special case of a prise
        // ..first get destSquare file :
        int file_ds = (ffsll(destSquare) - 1) % 8;

        if (bs->color) {
            if (file_is < file_ds) initSquare = destSquare << 7;
            else initSquare = destSquare << 9;
        }
        else {
            if (file_is < file_ds) initSquare = destSquare >> 9;
            else initSquare = destSquare >> 7;
        }
    }

    else {
        // normal moving (maybe two squares)
        if (bs->color) { // white (color==1)
            initSquare = destSquare << 8;
            if (bs->pawns[1] & initSquare) return initSquare;
            return (initSquare << 8); // 2 squares move
        }

        else { // black (color==0)
            initSquare = destSquare >> 8;
            if (bs->pawns[0] & initSquare) return initSquare;
            return (initSquare >> 8); // 2 squares move
        }
    }

    return initSquare;
}

/* other pieces : no flags, but harder to find the right init square..
 * types are here : 1=rooks, 2=knights, 3=bishops, 4=queens.
 * initial square of king is usually not so hard, so not included there.
 * rank_is and file_is are optional arguments giving hint for initial square, for moves like Nfg7
 */
Bitboard getInitSquare(BoardState* bs, Bitboard destSquare, char pieceType, int file_is, int rank_is) {
    Bitboard initSquare=0, firstBit_pp=0, lastBit_pp=0, piecesPos=0;
    int file_ds=0, rank_ds=0, position_ds=0, sizeShift=0;
    int color = bs->color;

    // ..first convert destSquare from Bitboard to bidimensional position (file, rank) :
    position_ds = ffsll(destSquare) - 1;
    file_ds = position_ds % 8;
    rank_ds = 7 - position_ds / 8;

	/* General method is :
	 *    1) reduce research field by looking if file_is or rank_is is given
	 *    2) look if the theoretically available pieces are only one ; in this case, quick answer
	 *    3) unlucky, check for every possible ways
	 */

    switch (pieceType) {


    case 'R': // rooks :

        // 1) :
        if (file_is >= 0) {
            piecesPos = bs->rooks[color] & files[file_is];
            if (file_is == file_ds) sizeShift = 8;
            else sizeShift = 1;
        }
        else if (rank_is >= 0) {
            piecesPos = bs->rooks[color] & ranks[rank_is];
            if (rank_is == rank_ds) sizeShift = 1;
            else sizeShift = 8;
        }

        else {
            // we don't know if file or rank ; will determine this
            // try 1) again
            if ( ! (bs->rooks[color] & ranks[rank_ds]) ) { // no rooks on destination rank
                piecesPos = bs->rooks[color] & files[file_ds];
                sizeShift = 8;
            }
            else if ( ! (bs->rooks[color] & files[file_ds]) ) { // no rooks on destination file
                piecesPos = bs->rooks[color] & ranks[rank_ds];
                sizeShift = 1;
            }
            else {
                // unlucky --> potential 3)
                piecesPos = bs->rooks[color] & (files[file_ds] | ranks[rank_ds]);
            }
        }

		// 2) try shortcut :
        firstBit_pp = ffsll(piecesPos) - 1;
        lastBit_pp = flsll(piecesPos) - 1;
        if (firstBit_pp == lastBit_pp) {
            // very easy : only one rook (in fact most common case)
            return (piecesPos);
        }

		/* special shortcut for rooks on a file or rank :
		 * if the "last" one is above the first bit set,
		 * or if the "first" one (palview board ordering) is below last bit,
		 * then we know for sure the initial square
		 */
        if (sizeShift) {
            if (position_ds < firstBit_pp) {
                // just return first bit of piecesPos :
                return ((Bitboard)1 << firstBit_pp);
            }
            if (position_ds > lastBit_pp) {
                // ..or last bit of piecesPos :
                return ((Bitboard)1 << lastBit_pp);
            }
            // unlucky case : "light" 3) (only two ways)
            return (tryTwoWays(bs, destSquare, piecesPos, 1,sizeShift,0,sizeShift));
        }

        // 3) "hard" case : we do a basic search from destination square..
        // luckily, will not happen so often (2+ rooks targeting the same square)
        // try up :
        Bitboard tmp = tryTwoWays(bs, destSquare, piecesPos, 1,8,0,8);
        if ( ~tmp ) return tmp;
        // ..and down :
        return (tryTwoWays(bs, destSquare, piecesPos, 1,1,0,1));


    case 'N': // knights :

        // first, apply right shift of stored bitboard from destSquare :
        piecesPos = knightsMask[file_ds];
        if (rank_ds < 4) piecesPos <<= (8 * (4 - rank_ds));
        else if (rank_ds > 4) piecesPos >>= (8 * (rank_ds - 4));
        piecesPos &= bs->knights[color]; // potential knights initial squares

		// 1) :
        if (file_is >= 0) {
            // we know that the moving knight is on this file
            // warning : there could still be 2 knights ! one can be pinned..
            piecesPos &= files[file_is];
        }
        else if (rank_is >= 0) {
            piecesPos &= ranks[rank_is];
        }

		// 2) :
        firstBit_pp = ffsll(piecesPos) - 1;
        lastBit_pp = flsll(piecesPos) - 1;
        if (firstBit_pp == lastBit_pp) return piecesPos;

        if (file_is >= 0 || rank_is >= 0) {
            // light 3) : there are two knights (never more !) in conflict
            // quite rare situation (one is pinned)
            initSquare = (Bitboard)1 << firstBit_pp;
            if (!checkCheck(bs, initSquare, destSquare)) return initSquare;
            return ((Bitboard)1 << lastBit_pp);
        }

        // 3) general case : there can be up to 8 knights
        // max. 2 of them could be pinned, but which ones ??!
        // method = try every possibility...
        int accSum = 0;
        while (piecesPos != 0) {
            firstBit_pp = ffsll(piecesPos);
            initSquare = (Bitboard)1 << (accSum + firstBit_pp - 1);
            if (!checkCheck(bs, initSquare, destSquare)) return initSquare;
            piecesPos >>= firstBit_pp;
            accSum += firstBit_pp;
        }
        return 1; // will (should..) never happen !

    case 'B': ; // bishops :

        // first, apply right shift of stored bitboards from destSquare :
        Bitboard mainDiagMask = diagonals[0];
        if (file_ds < rank_ds) {
            // upper part : delete low bits
            mainDiagMask >>= 8 * (rank_ds - file_ds);
        }
        else if (file_ds > rank_ds) {
            // lower part : delete high bits
            mainDiagMask <<= 8 * (file_ds - rank_ds);
        }

        Bitboard otherDiagMask = diagonals[1];
        int mRank_ds = 7 - rank_ds;
        if (file_ds < mRank_ds) {
            // lower part
            otherDiagMask <<= 8 * (mRank_ds - file_ds);
        }
        else if (file_ds > mRank_ds) {
            // upper part
            otherDiagMask >>= 8 * (file_ds - mRank_ds);
        }
        piecesPos = bs->bishops[color] & (mainDiagMask | otherDiagMask);

        // 1) : two cases where there can be max. 2 bishops answering conditions
        if (file_is >= 0) piecesPos = piecesPos & files[file_is];
        else if (rank_is >= 0) piecesPos = piecesPos & ranks[rank_is];

        // 2) : shortcut
        firstBit_pp = ffsll(piecesPos);
        lastBit_pp = flsll(piecesPos);
        if (firstBit_pp == lastBit_pp) {
            // easy : only one bishop == most common case for known file or rank
            return (piecesPos);
        }

        if (file_is >= 0) {
            // ..oops a bit less easy, really 2 bishops ; which one to choose ?
            if (file_is < file_ds) {
                // 0 == rightShift (1== leftShift)
                return (tryTwoWays(bs, destSquare, piecesPos, 1, 7, 0, 9));
            }
            return (tryTwoWays(bs, destSquare, piecesPos, 1, 9, 0, 7));
        }

        if (rank_is >= 0) {
            // which one to choose (again !) ?
            if (rank_is < rank_ds) {
                // 0 == rightShift (1== leftShift)
                return (tryTwoWays(bs, destSquare, piecesPos, 1, 7, 1, 9));
            }
            return (tryTwoWays(bs, destSquare, piecesPos, 0, 7, 0, 9));
        }

        // 3) "hard" case : we do a basic search from destination square..
        // = the most common case for bishops ! but generally only a few bishops on board..
        Bitboard tmpB = tryTwoWays(bs, destSquare, piecesPos, 1,7,1,9);
        if ( ~tmpB ) return tmpB;
        return (tryTwoWays(bs, destSquare, piecesPos, 0,7,0,9));

    case 'Q': // queens : rook + bishop

		// 1) : the case of a move like Qdd8 is quite rare, but has to be checked
        if (file_is >= 0) {
            piecesPos = bs->queens[color] & files[file_is];
            if (file_is == file_ds) sizeShift = 8; // not costly.. has to be tried !
        }
        else if (rank_is >= 0) {
            piecesPos = bs->queens[color] & ranks[rank_is];
            if (rank_is == rank_ds) sizeShift = 1;
        }

        else {
            // we don't know if file or rank ; will apply all masks
            // go for diagonals masks :
            Bitboard mainDiagMask = diagonals[0];
            if (file_ds < rank_ds) {
                // upper part : delete low bits
                mainDiagMask >>= 8 * (rank_ds - file_ds);
            }
            else if (file_ds > rank_ds) {
                // lower part : delete high bits
                mainDiagMask <<= 8 * (file_ds - rank_ds);
            }

            Bitboard otherDiagMask = diagonals[1];
            int mRank_ds = 7 - rank_ds;
            if (file_ds < mRank_ds) {
                // lower part
                otherDiagMask <<= 8 * (mRank_ds - file_ds);
            }
            else if (file_ds > mRank_ds) {
                // upper part
                otherDiagMask >>= 8 * (file_ds - mRank_ds);
            }
            piecesPos = bs->queens[color] & (files[file_ds] | ranks[rank_ds] | mainDiagMask | otherDiagMask);
        }

		// 2) : try shortcut
        firstBit_pp = ffsll(piecesPos) - 1;
        lastBit_pp = flsll(piecesPos) - 1;
        if (firstBit_pp == lastBit_pp) {
            // very easy : only one queen ! ==(really) most common case
            return (piecesPos);
        }

        if (sizeShift) {
            // we know the file or rank ; easy : same as rooks
            if (position_ds < firstBit_pp) {
                // we just return first bit of piecesPos :
                return ((Bitboard)1 << firstBit_pp);
            }
            if (position_ds > lastBit_pp) {
                // ..or last bit of piecesPos :
                return ((Bitboard)1 << lastBit_pp);
            }
            // "light" 3) : only two ways
            return (tryTwoWays(bs, destSquare,piecesPos, 1,sizeShift,0,sizeShift));
        }

        // 3) : "hard" case : we do a basic search from destination square,
        // in every direction. this is the most common case when 2+ queens (...)
        Bitboard tmpQ = tryTwoWays(bs, destSquare, piecesPos, 1,8,0,8);
        if ( ~tmpQ ) return tmpQ;
        tmpQ = tryTwoWays(bs, destSquare, piecesPos, 1,1,0,1);
        if ( ~tmpQ ) return tmpQ;
        tmpQ = tryTwoWays(bs, destSquare, piecesPos, 1,7,1,9);
        if ( ~tmpQ ) return tmpQ;
        return (tryTwoWays(bs, destSquare, piecesPos, 0,7,0,9));

    default: return 0;
    }
}
