#define _GNU_SOURCE //for ffsll
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "types.h"
#include "init_square.h"
#include "move.h"
#include "safe_alloc.h"
#include "stack_sm.h"

// a piece code for flags encoding of pawns promotions,
// and for some checks in main loop :
int pieceCode(char piece) {
    switch (piece) {
    case 'R': return 1;
    case 'N': return 2;
    case 'B': return 3;
    case 'Q': return 4;
    case 'K': return 5;
    }
    return 0; // will not happen
}

// for the very rare case of three/four targeting sliders (non-standard SAN, but..),
// we add the piece name to be able to draw the fig. symbol on HTML ;
// e.g. : g8e6 for "Qg8e6", with three queens e8, g8, g6.
char invPieceCode(int code) {
	switch (code) {
	case 1: return 'R';
	case 3: return 'B';
	case 4: return 'Q';
	}
	return 0; // will not happen
}

// "inverse ffsll" function : find the Bitboard from (file, rank) coordinates
Bitboard getBitPos(char file, char rank) {
    return ( (Bitboard)1 << ( (file - 'a') + 8 * (8 - (rank - '0')) ) );
}


/**********************************************************************************
 * This is the core function : take a string like Nh6 in parameter,
 * and return initial square + flags for the move
 * flags == two ASCII char's from 64 (@) to 126 (~) ; two highest bits are always 01
 * first flag :
 *    3 lowest bits indicate the type of capture piece OR the type of castle
 *    3 next bits indicate the type of promoted piece OR a "prise en passant"
 * second flag : encode depth of current move into (possibly) nested variations ;
 *               up to 14 nested depths + main line allowed ; also 2 bits
 *               to say if this move is var. beginning, and maybe continuation
 *********************************************************************************/
Move* getMoveFromString(BoardState* bs, char* moveStr, int* beginVar, int* continuation, int depthVar) {
    Bitboard initSquare=0, destSquare=0;
    Short flags = 0x40;

    switch (moveStr[0]) {

    case 'O':

        // castle move

        initSquare = ((Bitboard)16 << (bs->color * 56));
        if (moveStr[3] == '-') {
            // big castle
            destSquare = (initSquare >> 2);
            flags += 7;
        }
        else {
            // small castle
            destSquare = (initSquare << 2);
            flags += 6;
        }
        break;

    case 'a':
    case 'b':
    case 'c':
    case 'd':
    case 'e':
    case 'f':
    case 'g':
    case 'h':

        // (most often) a pawn move :

		if (moveStr[2] == '=' || pieceCode(moveStr[2])) {
			// promotion, maybe with wrong notation ; process all anyway
			destSquare = getBitPos(moveStr[0],moveStr[1]);
			if (bs->color) initSquare = destSquare << 8;
            else initSquare = destSquare >> 8;
			if (moveStr[2] == '=') flags += (pieceCode(moveStr[3]) << 3);
			else flags += (pieceCode(moveStr[2]) << 3);
		}
        else if (moveStr[1] == 'x') {
            // a pawn takes something ; we have to flag accordingly
            destSquare = getBitPos(moveStr[2],moveStr[3]);
            initSquare = getInitSquare_pawns(bs, destSquare, 1, moveStr[0] - 'a');
            int captured = whichPiece(bs, destSquare, 1 - bs->color);
            if (!captured) {
                // special case : prise en passant
                flags += 0x28;
            }
            else {
                flags += captured;
                if (moveStr[4] == '=') {
                    // a pawn promotion which is also a capture :
                    flags += (pieceCode(moveStr[5]) << 3);
                }
            }
        }
        else if (moveStr[2] >= 'a' && moveStr[2] <= 'h') {
            /* the very very special case of three or four sliders all
             * targeting the same square (I've never seen that but..) ;
             * in this case the (possible) capture mark 'x' could not be printed,
             * nor is the piece's type (we find it in the function
             * "invPieceCode" above for later figurines printing in HTML).
             */
            initSquare = getBitPos(moveStr[0],moveStr[1]);
            destSquare = getBitPos(moveStr[2],moveStr[3]);
            flags += whichPiece(bs, destSquare, 1 - bs->color);
        }
        else {
            // simple moving like 1.e4
            destSquare = getBitPos(moveStr[0],moveStr[1]);
            initSquare = getInitSquare_pawns(bs, destSquare, 0, 0);
        }
        break;

    case 'R':
    case 'N':
    case 'B':
    case 'Q':
    case 'K': ;

        // a piece is moving and maybe taking something

        int nextRead = 1;
        int file_ds = -1;
        int rank_ds = -1;
        if (moveStr[1] >= '1' && moveStr[1] <= '8') {
            // rank is specified
            rank_ds = moveStr[1] - '1';
            nextRead++;
        }
        else if (moveStr[2] == 'x' || (moveStr[1] >= 'a' && moveStr[1] <= 'h'
				&& moveStr[2] >= 'a' && moveStr[2] <= 'h')) {
            // file is specified :
            file_ds = moveStr[1] - 'a';
            nextRead++;
        }

        // nextRead char is (most often) either 'x' for takes, or (file, rank) coordinates
        if (moveStr[nextRead] == 'x') {
            destSquare = getBitPos(moveStr[nextRead+1],moveStr[nextRead+2]);
            flags += whichPiece(bs, destSquare, 1 - bs->color);
        }
        else {
            if (nextRead == 1 && (moveStr[3] == 'x' || (moveStr[3] >= 'a' && moveStr[3] <= 'h'))) {
				// the very special case of 3--4 targeting sliders, in SAN notation (e.g. Qd5xd7+)
				initSquare = getBitPos(moveStr[1],moveStr[2]);
				if (moveStr[3]=='x') {
					destSquare = getBitPos(moveStr[4],moveStr[5]);
					flags += whichPiece(bs, destSquare, 1 - bs->color);
				}
				else {
					destSquare = getBitPos(moveStr[3],moveStr[4]);
				}
				break;
			}
            else {
				// simple move without prise
				destSquare = getBitPos(moveStr[nextRead],moveStr[nextRead+1]);
			}
        }
        if (moveStr[0] == 'K') initSquare = bs->kings[bs->color]; // king moves are easy to trace..
        else initSquare = getInitSquare(bs, destSquare, moveStr[0], file_ds, rank_ds);
        break;
    }

    Move* mv = safe_malloc (sizeof(Move));
    mv->vdepth = 0x40 + depthVar; // limit depth == 14 because 15 would lead to DEL (unprintable)
	if (*beginVar) mv->vdepth |= 0x20;
    if (*continuation) mv->vdepth |= 0x10;
	*beginVar = 0;
	*continuation = 0;
	mv->initSquare = initSquare;
    mv->destSquare = destSquare;
    mv->flags = flags;
    return mv;
}


/****************************************************************************
 * Another key function, which send a token like "3...", "Ba6",
 * or parenthesis for variations (most often to) to the string of "printable" moves
 ***************************************************************************/
void sendTokenTo(char** gameString, char* tok, char pieceC) {
	// pieceC if special case of three or four targeting sliders : e.g. a8c6

    int i, shiftInit=0, sizeTok = strlen(tok);
	if (pieceC) shiftInit = 1; // we will add printing like 'R', 'B' or 'Q'
	//strings initialized to 0, so, never NULL here
    int initPos = strlen(*gameString);
    *gameString = safe_realloc(*gameString, initPos + shiftInit + sizeTok + 1);

	if (shiftInit) (*gameString)[initPos] = pieceC;
    for (i = 0; i < sizeTok; i++)
        (*gameString)[initPos + shiftInit + i] = tok[i];
    (*gameString)[initPos + shiftInit + sizeTok] = 0;
}


/*****************************************************************************
 * This function transform a move (structure) into 4 ASCII chars
 * the first two are initial and destination squares, the two last are the two flags
 * example :
 *    according to the board coordinates ordering, 10 --> 26 represents c7 to c5 ;
 *    let's imagine this is a capture of a bishop, at variation depth 2
 *    move will be ASCII(63 + 10) --> ASCII(63 + 26) == 'I' --> 'Y'
 *    and flags 01 000 011 [capture bishop] == 'C', and 01 000 010 [depth 2] == 'B'
 *    ==> the move 4 characters' string is "IYCB" ; simple and efficient !
 *****************************************************************************/
void moveToASCII(char** asciiMoves, Move* mv) {
	// WARNING : we have to escape backslash character with another one,
    // otherwise it will be misunterstood by javascript !
	int expandMem = 0;
	int pos0=0, pos1=0, pos2=0, pos3=0;
	if (mv->initSquare & 0x0000000020000000) { // backslash code on our board
		expandMem++;
		pos0 = 1;
	}
	if (mv->destSquare & 0x0000000020000000) {
		expandMem++;
		pos1 = 1;
	}
	if (mv->flags == 0x5C) {
		expandMem++;
		pos2 = 1;
	}
	if (mv->vdepth == 0x5C) {
		expandMem++;
		pos3 = 1;
	}

    int initPos = strlen(*asciiMoves), addPos = 0;
    //strings initialized to 0, so, never NULL here
	*asciiMoves = safe_realloc(*asciiMoves, initPos + 5 + expandMem);
    (*asciiMoves)[initPos] = ffsll(mv->initSquare) + 62;
	if (pos0) {
		(*asciiMoves)[initPos + 1] = (*asciiMoves)[initPos];
		addPos++;
	}

    (*asciiMoves)[initPos + 1 + addPos] = ffsll(mv->destSquare) + 62;
    if (pos1) {
		(*asciiMoves)[initPos + 2 + addPos] = (*asciiMoves)[initPos + 1 + addPos];
		addPos++;
	}

    (*asciiMoves)[initPos + 2 + addPos] = mv->flags;
    if (pos2) {
		(*asciiMoves)[initPos + 3 + addPos] = (*asciiMoves)[initPos + 2 + addPos];
		addPos++;
	}

    (*asciiMoves)[initPos + 3 + addPos] = mv->vdepth;
    if (pos3) {
		(*asciiMoves)[initPos + 4 + addPos] = (*asciiMoves)[initPos + 3 + addPos];
		addPos++;
	}
    (*asciiMoves)[initPos + 4 + addPos] = 0;
}


/****************************************************************
 * High-level function to print the current move
 * to the string of "printable" moves ; it also print the ASCII string
 * of a move to the string "asciiMoves" ; see funcs. above
 ***************************************************************/
void processMove(StateAndMove* sm, char** printableMoves, char** asciiMoves,
                 char** gameStr, int* beginVar, int* continuation, int depthVar) {

    // a move is divided into at most 4 tokens :
    // numbers and dots (x2, e.g. 2 ... c5), move itself, and optional nag.

    // [optional] FIRST TOKEN : dots or number (imagine 19 ...Rf6, 12. Ra4)
    // black moves have this token only when exiting a variation [e.g. "[...] 2.d4 f5 ) 2...a6"]
    char* pchar = *gameStr;
    while (*pchar=='.' || (*pchar >= '0' && *pchar <= '9')) pchar++;
    char saveChar = *pchar;
    *pchar = 0;
    if (pchar != *gameStr) sendTokenTo(printableMoves, *gameStr, 0);
    *pchar = saveChar;

    // SECOND TOKEN : could be dots again ; think about 3. ..e5 (messy typing, but allowed)
	while (*pchar==' ' || *pchar == '\n' || *pchar == '\r' || *pchar == '\t') pchar++;
	char* endTok = pchar;
	while (*endTok == '.') endTok++;
	saveChar = *endTok;
    *endTok = 0;
	if (endTok != pchar) sendTokenTo(printableMoves, pchar, 0);
	*endTok = saveChar;
	pchar = endTok;

	// "THIRD" TOKEN : the move itself
    while (*pchar==' ' || *pchar == '\n' || *pchar == '\r' || *pchar == '\t') pchar++;
    endTok = pchar;
    while (*endTok != ' ' && *endTok != '\n' && *endTok != '\r' && *endTok != '\t' && *endTok != '('
                          && *endTok != ')' && *endTok != '[' && *endTok != ']' && *endTok != '{')
        endTok++;
    saveChar = *endTok;
    *endTok = 0;
	if (pchar[0] >= 'a' && pchar[0] <= 'h' && pchar[1] >= '1' && pchar[1] <= '8' && pchar[2] >= 'a' && pchar[2] <= 'h') {
		// very special case of three/four targeting sliders : we print piece character
		Bitboard initSquare = getBitPos(pchar[0],pchar[1]);
		char pieceC = invPieceCode( whichPiece(sm->bs, initSquare, sm->bs->color) - 1 );
    	sendTokenTo(printableMoves, pchar, pieceC);
	}
	else sendTokenTo(printableMoves, pchar, 0);

    // hard task : get full move from PGN-like move :
    Move* mv = getMoveFromString(sm->bs, pchar, beginVar, continuation, depthVar);
    moveToASCII(asciiMoves,mv);
    *endTok = saveChar;

    // update move in structure 'sm' (which describes current internal board state) :
    if (sm->mv != NULL) free(sm->mv);
    sm->mv = mv;

    // [optional] "FOURTH" TOKEN : the nag $xx ; another nag can follow, but will be treated in main loop
    pchar = endTok;
    while (*pchar==' ' || *pchar == '\n' || *pchar == '\r' || *pchar == '\t') pchar++;
    if (*pchar == '$') {
        endTok = pchar+1;
        while (*endTok >= '0' && *endTok <= '9') endTok++;
        saveChar = *endTok;
        *endTok = 0;
        sendTokenTo(printableMoves, pchar, 0);
        *endTok = saveChar;
    }

    // set game string to the next character after move :
    *gameStr = endTok;

    //end by sending token SPACE, for breeze..
    sendTokenTo(printableMoves, " ", 0);
}

// Set a BoardState object the the initial board set-up, whites to move
// color 0 == Black, color 1 == White
void initializeBoard(BoardState* bs) {
    bs->fullBoard = 0xFFFF00000000FFFF;
    bs->pawns = safe_malloc (2*sizeof(Bitboard));
    bs->pawns[1] = 0x00FF000000000000;
    bs->pawns[0] = 0x000000000000FF00;
    bs->rooks = safe_malloc(2*sizeof(Bitboard));
    bs->rooks[1] = 0x8100000000000000;
    bs->rooks[0] = 0x0000000000000081;
    bs->knights = safe_malloc(2*sizeof(Bitboard));
    bs->knights[1] = 0x4200000000000000;
    bs->knights[0] = 0x0000000000000042;
    bs->bishops = safe_malloc(2*sizeof(Bitboard));
    bs->bishops[1] = 0x2400000000000000;
    bs->bishops[0] = 0x0000000000000024;
    bs->queens = safe_malloc(2*sizeof(Bitboard));
    bs->queens[1] = 0x0800000000000000;
    bs->queens[0] = 0x0000000000000008;
    bs->kings = safe_malloc(2*sizeof(Bitboard));
    bs->kings[1] = 0x1000000000000000;
    bs->kings[0] = 0x0000000000000010;
    bs->color = 1; // 1 means "white to move"
}


/**********************************************************************************************
 * MAIN LOOP : MOVE ANALYZER
 * argv[1] is the string of the game "PGN-like"
 * Method :
 *    + we keep a stack of StateAndMove to keep track of states where we dive into variations,
 *      to be able to return from them at their end.
 *    + when we meet a '('", we push BoardState and Move
 *    + when we meet a ')', we pop BoardState and Move
 *    + when meet a simple move : process ;
 *    + when meet a comment : skip ==> will be processed in PHP program
 * Nags : printed as they are, further transformations in PHP code.
 **********************************************************************************************/
int main (int argc, char** argv) {

    //printableMoves : the human-readable string essentially for printing
    //asciiMoves : the "palview like" string for javascript process
    char *printableMoves, *asciiMoves, *pchar = argv[1];
    printableMoves = safe_malloc (1);
    printableMoves[0] = 0;
    asciiMoves = safe_malloc (1);
    asciiMoves[0] = 0;

    // initialize board :
    BoardState* bs = safe_malloc (sizeof(BoardState));
    initializeBoard(bs);
    StateAndMove* sm = safe_malloc (sizeof(StateAndMove)); // current board state + move about to be played. useful if variations !
    sm->bs = bs;
	sm->mv = NULL;
    StackSM* smStack = safe_malloc (sizeof(StackSM)); // stack of StateAndMove's, useful if variations..
    smStack->sm = NULL;

	int beginVar = 0; // tells if next move marks a variation beginning    
	int continuation = 0; // tells if next move will be a continuation
	StackInt* cbStack = safe_malloc (sizeof(StackInt)); // linked list to stack continuation bits
	cbStack->cont = NULL;
    int depthVar = 0; // variations depth (0 = main line ; 1 = first level of analysis... up to 14)

    while (1) {

        // find the next non space character
        while (*pchar == ' ' || *pchar == '\n' || *pchar == '\r' || *pchar == '\t') pchar++;
        
        // a few options only : a move number, '(' or ')' for opening / closing variations and '{' for opening comment
        // also possible : prefix-nag, continuation symbol, null-move, and.. game ending !

        if (*pchar == '*' || *pchar == '0') {
            // GAME ENDING : we already know the game result by parsing in PHP file ;
            // so, redondant... we don't even read it [here, 0-1 or * for "unfinished"]
			int lastPchar = strlen(printableMoves) - 1;
			printableMoves[lastPchar] = 0;
            printf("%s*%s\n",printableMoves,asciiMoves);
			printableMoves[lastPchar] = '@';
            free(printableMoves);
            free(asciiMoves);
            free_contentsSM(sm);
			free(sm);
            free(smStack);
			free(cbStack);
            return 0;
        }

        if (pchar[0] == '1' && (pchar[1]=='-' || pchar[1]=='/')) {
            // game ending again (1-0 or 1/2-1/2)
			int lastPchar = strlen(printableMoves) - 1;
			printableMoves[lastPchar] = 0;
            printf("%s*%s\n",printableMoves,asciiMoves);
			printableMoves[lastPchar] = '@';
            free(printableMoves);
            free(asciiMoves);
            free_contentsSM(sm);
			free(sm);
            free(smStack);
			free(cbStack);
            return 0;
        }

        /* first character of a game move
         * it can be : a digit 1-9 if number provided,
         * a letter a-h if pawn move with no number,
         * a dot to indicate black move,
         * RNBQK for black piece move, or O for castle
         */
        if ((*pchar >= '1' && *pchar <= '9') || (*pchar >= 'a' && *pchar <= 'h')
                || pieceCode(*pchar) || *pchar == 'O' || *pchar == '.') {
            processMove(sm, &printableMoves, &asciiMoves, &pchar, &beginVar, &continuation, depthVar);
            play(sm->bs, sm->mv); // changes board state sm->bs
        }

		else if (*pchar == '$') {
			// very special case of an untreated nag : this is a prefix one, like "with the idea of.."
			// it could occur if a move is already nagged, and another nag follow, like e4 $1 $140 in a variation..
			// so we just send it to printable moves, and it will appear as a move prefix !
			char* endTok = pchar+1;
			while (*endTok >= '0' && *endTok <= '9') endTok++;
			char saveChar = *endTok;
			*endTok = 0;
			sendTokenTo(&printableMoves, pchar, 0);
			sendTokenTo(&printableMoves, " ", 0);
			*endTok = saveChar;
			pchar = endTok;
		}

        else if (*pchar == '(' || *pchar == '[') {
			// beginning of variation : we have to remember it
			beginVar = 1;
            sendTokenTo(&printableMoves, "( ", 0);
			if (pchar[1] == '*') {
				continuation = 1;
				pchar++;
			}
            else takeback(sm->bs, sm->mv); // if no cont., first move in a variation is a takeback of last move played
            pushSM(&smStack, safe_copySM(sm)); // we have to stack it to get back here later
			int* contBit = safe_malloc (sizeof(int));
			*contBit = continuation;
			pushInt(&cbStack, contBit); // also stack continuation bit needed when we get back here
            depthVar++;
            pchar++;
        }

        else if (*pchar == ')' || *pchar == ']') {
            sendTokenTo(&printableMoves, ") ", 0);
            free_contentsSM(sm);
			free(sm);
            sm = popSM(&smStack);
			int* contBit = popInt(&cbStack);
			if ( ! *contBit ) play(sm->bs, sm->mv); // play again the stored move when entered variation, IF NEEDED
			free(contBit);
            depthVar--;
            pchar++;
        }

		else if (*pchar == '<' || *pchar == '-') {
			// null move = very special case.
			// we send it directly to the ascii game string (nothing to print !)
			Short vdepth = 0x40 + depthVar;
			if (beginVar) vdepth |= 0x20;
    		if (continuation) vdepth |= 0x10;
			char nullMoveString[] = "??@ \0";
			nullMoveString[3] = vdepth;
			sendTokenTo(&asciiMoves, nullMoveString, 0);
			beginVar = 0;
			continuation = 0;
			sm->bs->color = 1 - sm->bs->color; // WARNING : color changes during a null move !
			pchar += 2;
		}

		// we completely ignore comments, because they will be parsed and transformed in PHP program !
        else if (*pchar == '{') {
            // run through until '}'
            while (*pchar != '}') pchar++;
        }

		else pchar++; // only left option should be '}'..
    }

    return 0; // never reached
}

